/*
 * Controller Aera Network(CAN) driver for Linux 2.6
 * 
 * Re-written by Zumeng.Chen <zumeng.chen@windriver.com>
 * Based on peak-linux-driver.3.17.tar.gz.
 * 
 * This file is licensed under the terms of the GNU General Public
 * License version 2. This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/device.h>
#include <asm/atomic.h>
#include <asm/types.h>

#include "can-base.h"

/* ------------------------------------------------------------------------ */
#define FIFOSIZE 200
/* Used for minor number allocation */
static struct can_endpoint_t *can_endpoint[CAN_MAX_ENDPOINTS];
struct class *can_dev_class = NULL;

struct can_endpoint_t *can_endpoint_get_by_minor(unsigned index)
{
	if (index < 0 || index >= CAN_MAX_ENDPOINTS)
		return NULL;
	return can_endpoint[index];
}

struct can_endpoint_t * can_endpoint_init(int factorTx2Rx)
{
	struct can_endpoint_t *endpoint = (struct can_endpoint_t *)kmalloc(sizeof(struct can_endpoint_t), GFP_KERNEL);

	memset((void *) endpoint, 0x0, sizeof(struct can_endpoint_t));
	
	if (can_fifo_init(&endpoint->txfifo, FIFOSIZE)) {
		goto free_exit;
	}
	printk(KERN_INFO "CAN FIFO: %dBytes for Tx\n", FIFOSIZE* (sizeof(struct can_frame)));
	
	if (!(can_fifo_init(&endpoint->rxfifo, FIFOSIZE*factorTx2Rx))) {
		goto normal_exit;
	}

	printk(KERN_ERR "CAN: %s out of memory\n",__func__);
	kfree((const void *) endpoint->rxfifo.data);
free_exit:
	kfree((const void *) endpoint);
normal_exit:	
	return endpoint;
}

void can_endpoint_free(struct can_endpoint_t *endpoint)
{
	kfree((const void *) endpoint->txfifo.data);
	kfree((const void *) endpoint->rxfifo.data);
	kfree((const void *) endpoint);
	return ;
}

/* read a frame from buffer-FIFO */
can_frame_t * can_endpoint_readframe(struct can_endpoint_t *endpoint, int uid, int  *retval)
{
	can_frame_t  * frame = NULL;

	can_fifo_lock(&endpoint->rxfifo);
	frame = can_fifo_pop(&endpoint->rxfifo, uid, retval);
	can_fifo_unlock(&endpoint->rxfifo); 
	/*Try to release lock asap. */

	return frame;
}

/* Try to write a CAN frame from buffer-FIFO to Hard-Tx */
void  can_endpoint_writeframe(struct can_endpoint_t *endpoint) 
{
	int sending = 0;

	can_fifo_lock(&endpoint->txfifo);
	sending = atomic_sub_and_test(1, &endpoint->send_ready_flag);
	can_fifo_unlock(&endpoint->txfifo);	

	if (sending)
		endpoint->initiate2write(endpoint);
	return;
}

/* Try to get free frame from buffer-FIFO */
can_frame_t * can_endpoint_getframe(struct can_endpoint_t *endpoint, int * retval) 
{
	can_frame_t * frame = NULL;

	can_fifo_lock(&endpoint->txfifo);
	frame = can_fifo_get(&endpoint->txfifo, retval);
	can_fifo_unlock(&endpoint->txfifo);	

	return frame;
}

int can_endpoint_adduser(struct can_endpoint_t *endpoint)
{
	int uid;

	can_fifo_lock(&endpoint->rxfifo);
	uid = can_fifo_adduser(&endpoint->rxfifo);
	can_fifo_unlock(&endpoint->rxfifo);
	
	return uid;
}

int can_endpoint_deluser(struct can_endpoint_t *endpoint, int uid)
{
	int retval = 0;

	can_fifo_lock(&endpoint->rxfifo);
	retval = can_fifo_deluser(&endpoint->rxfifo, uid);
	can_fifo_unlock(&endpoint->rxfifo);
	
	return retval;
}

/*  This function will be called by chip-specific implementations; return minor number  */
int can_endpoint_register(struct can_endpoint_t *endpoint) {
	int retval, i, minor = -1;

	/* Allocate a minor number */
	for (i=0; i<CAN_MAX_ENDPOINTS; i++) {
		if (can_endpoint[i] == NULL) {
			minor = i;
			break;
		}
	}

	if (minor == -1) {
		printk(KERN_ERR "CAN: Ran out of minor numbers!\n");
		return -EFAULT;
	}

	can_endpoint[minor] = endpoint;
	endpoint->minor = minor;
	endpoint->txfifo.end[0] = endpoint->txfifo.start;

	init_waitqueue_head(&endpoint->write_queue);
	init_waitqueue_head(&endpoint->read_queue);

	snprintf(endpoint->name, 7, "can%i", minor);
	retval = minor;

	printk(KERN_INFO "CAN: Registering CAN endpoint %s\n", endpoint->name);

	/* To create a class device */
	class_device_create(can_dev_class, NULL, MKDEV(CAN_MAJOR, minor), NULL, endpoint->name);

	return minor;
}

/* Unregister a endpoint (device specific module unloaded or USB hotplug) */
void can_endpoint_unregister(struct can_endpoint_t *endpoint) 
{
	class_device_destroy(can_dev_class, MKDEV(CAN_MAJOR, endpoint->minor));

	can_endpoint[endpoint->minor] = NULL;
	can_endpoint_free(endpoint);
}

EXPORT_SYMBOL(can_endpoint_free);
EXPORT_SYMBOL(can_endpoint_register);
EXPORT_SYMBOL(can_endpoint_init);
EXPORT_SYMBOL(can_endpoint_unregister);
