/*
 * Controller Aera Network(CAN) driver for Linux 2.6
 * 
 * Re-written by Zumeng.Chen <zumeng.chen@windriver.com>
 * Based on peak-linux-driver.3.17.tar.gz.
 * 
 * This file is licensed under the terms of the GNU General Public
 * License version 2. This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/device.h>
#include <asm/atomic.h>
#include <asm/types.h>
#include <asm/uaccess.h>

#include "can-base.h"

/* --------------CAN File operation functions start --------------------*/
/* File operation function for CAN endpoint */
/* CAN Open function . */
static int can_endpoint_open(struct inode *inode, struct file *file) 
{
	struct can_endpoint_t *endpoint;
	int uid = -1, deviceId = iminor(inode);
	struct can_dev_client * can_client = kmalloc(sizeof(struct can_dev_client),
											GFP_KERNEL);
	
	endpoint = (struct can_endpoint_t *)can_endpoint_get_by_minor(deviceId);
	
	if (endpoint == NULL || can_client == NULL)
		return -ENODEV;

	if((endpoint->rxfifo.user_counter) == (u32)CAN_MAX_USER_BIT)
		return -EBUSY;
	
	can_client->endpoint = endpoint;

	uid = can_endpoint_adduser(endpoint);

	if (uid < 0) 
		return uid;
	
	can_client->uid = uid;
	can_client->magic_num = ((uid) ^ (unsigned int)endpoint);
	
	printk(KERN_INFO "CAN: Opening endpoint %s (uid %i)\n", endpoint->name, uid);

	file->private_data = (void *) can_client;

	return 0;
}

/* CAN release function */
static int can_endpoint_release(struct inode *inode, struct file *file) 
{
	struct can_dev_client * can_client = (struct can_dev_client *)file->private_data;
	struct can_endpoint_t *endpoint = can_client->endpoint;
	int uid = can_client->uid;
	int retval = -1;

	if(can_client->magic_num != ((uid) ^ (unsigned int)endpoint)) {
		printk(KERN_ERR"CAN can_client has been overrun by unknown reason.\n ");
	}
	
	printk(KERN_INFO "CAN: /dev/can%i closed\n", can_client->uid);

	retval = can_endpoint_deluser(endpoint, can_client->uid);

	kfree((const void *)can_client);
	if(retval != 0){
		printk(KERN_ERR "CAN: wicked error in %s \n",__func__);
		retval = -EFAULT;
	}
	
	return retval;
}

/* CAN read function*/
static ssize_t can_endpoint_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	struct can_dev_client * can_client = (struct can_dev_client *)file->private_data;
	struct can_endpoint_t *endpoint = can_client->endpoint;
	can_frame_t * frame = NULL;	
	int uid = can_client->uid;
	int retval = -1;
	size_t index = 0;

	if(can_client->magic_num != ((uid) ^ (unsigned int)endpoint)) {
		printk(KERN_ERR"CAN can_client has been overrun by unknown reason.\n ");
		return -EFAULT;
	}

	while (count - index >= sizeof(struct can_frame)) {
		
		frame = can_endpoint_readframe(endpoint, can_client->uid, &retval);
		
		if(retval == 0){
			retval = copy_to_user(buf, frame, sizeof(struct can_frame));	
			if (retval) 
				return -EFAULT;
			index += sizeof(struct can_frame);
			buf += sizeof(struct can_frame);
			continue;
		}
		
		if (retval == -ENODATA) {
			if (file->f_flags & O_NONBLOCK) {
				if (index == 0)
					return -EAGAIN;
				else
					return index;
			} else {
				interruptible_sleep_on(&endpoint->read_queue);
				if (signal_pending(current)) {
					if (index == 0)
						return -EINTR;
					return index;
				}
			}
		} else  {
			printk(KERN_ERR "%s: Unknow error when read from fifo retval = %d \n",__func__,retval);
			return retval;
		} 
	}

	return index;
}

/*CAN  write function */
static ssize_t can_endpoint_write(struct file *file, const char *buf, size_t count, loff_t *ppos) 
{
	struct can_dev_client * can_client = (struct can_dev_client *)file->private_data;
	struct can_endpoint_t *endpoint = can_client->endpoint;
	int uid = can_client->uid;
	int retval = -1;
	size_t index = 0;

	if(can_client->magic_num != ((uid) ^ (unsigned int)endpoint)) {
		printk(KERN_ERR"CAN can_client has been overrun by unknown reason.\n ");
		return -EFAULT;
	}
	
	while (count - index >= sizeof(struct can_frame)) {
		can_frame_t * frame = NULL;
		
		frame = can_endpoint_getframe(endpoint, &retval);
		
		if(retval == 0) {
			if (copy_from_user(frame, (can_frame_t *) &buf[index], sizeof(struct can_frame)))
				return -EFAULT;
		
			can_endpoint_writeframe(endpoint);
			index += sizeof(struct can_frame);
			continue;
		}

		if (retval == -EBUSY) {
			if (file->f_flags & O_NONBLOCK) {
				if (index == 0)
					return -EAGAIN;
				else
					return index;
			}
			interruptible_sleep_on(&endpoint->write_queue);
			if (signal_pending(current)) {
				if (index == 0)
					return -EINTR;
				return index;
			}
		} else if (retval < 0) {
			printk(KERN_ERR "%s: Unknow error when write to fifo retval = %d \n",__func__,retval);
			return retval;
		} 
	}

	return index;
}

/* CAN IOCTL function*/
static int can_endpoint_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg) 
{
	struct can_dev_client * can_client = (struct can_dev_client *)file->private_data;
	struct can_endpoint_t *endpoint = can_client->endpoint;
	int uid = can_client->uid;
	int retval = 0;

	if(can_client->magic_num != ((uid) ^ (unsigned int)endpoint)) {
		printk(KERN_ERR"CAN can_client has been overrun by unknown reason.\n ");
		return -EFAULT;
	}

	can_fifo_lock(&endpoint->rxfifo);
	can_fifo_lock(&endpoint->txfifo);
	switch (cmd) {
		case CAN_SETBAUDRATE:
			endpoint->set_bitrate(endpoint, arg);
			break;
		case CAN_IOCHIPRESET:
			if(endpoint->chip_reset != NULL)
				endpoint->chip_reset(endpoint);
			else
				retval = -EPERM;	
			break;	
		case CAN_MODE:
			if(endpoint->set_mode != NULL)
				endpoint->set_mode(endpoint,arg);
			else
				retval = -EPERM;			
			break;			
		default:
			retval = -EINVAL;
	}
	
	can_fifo_unlock(&endpoint->txfifo);
	can_fifo_unlock(&endpoint->rxfifo);

	if(retval == -EINVAL)
		printk(KERN_WARNING "CAN device: Invalid ioctl request.\n");
	
	return retval;
}

static struct file_operations can_endpoint_fops = {
	.owner   = THIS_MODULE,
	.open    = can_endpoint_open,
	.release = can_endpoint_release,
	.write   = can_endpoint_write,
	.read    = can_endpoint_read,
	.ioctl   = can_endpoint_ioctl
};
/* --------------CAN File operation functions end --------------------*/


/* -----------module function start ------------------------------- */

static int __init can_dev_init(void) 
{
	int err = 0;

	can_dev_class = class_create(THIS_MODULE, "can-dev");
	if (IS_ERR(can_dev_class)) {
		err = PTR_ERR(can_dev_class);
		goto err_exit;
	}

	/* Register can device operation to VFS*/
	if (register_chrdev(CAN_MAJOR, "can-dev", &can_endpoint_fops)) {
		printk(KERN_ERR "Register the CAN major number failed !\n");
		err = -EFAULT;		
		goto err_destroy_class;
	}
	
	printk(KERN_INFO "CAN device class register Version 0.1\n");
	return 0;
	
err_destroy_class:
	class_destroy(can_dev_class);
err_exit:
	return err;	
}

static void __exit can_dev_cleanup(void) {
	unregister_chrdev(CAN_MAJOR, "can-dev");
	class_destroy(can_dev_class);
}

module_init(can_dev_init);
module_exit(can_dev_cleanup);
/* -----------module function end ------------------------------- */

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("MPC52xx CAN Device class support");
