/*
 * Controller Aera Network(CAN) driver for Linux 2.6
 * 
 * Re-written by Zumeng.Chen <zumeng.chen@windriver.com>
 * Based on peak-linux-driver.3.17.tar.gz.
 * 
 * This file is licensed under the terms of the GNU General Public
 * License version 2. This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/device.h>
#include <asm/atomic.h>
#include <asm/types.h>

#include "can-fifo.h"

/* 
 * It is somewhat confusion, current definition is if only there is a valid
 * reader, this function return 0; otherwise a negative return value: 
 * This function will be called only if read interrupter happen.
 */
static int can_fifo_full_check(can_fifo_t *fifo) 
{
	int i,retval = 0;
#ifdef CONFIG_CAN_DEBUG
	int overrun_bit = fifo->rd_overrun_bit;
#endif 

	for (i=0; i<CAN_MAX_USER; i++) {
		if(fifo->end[i] != -1) {
			if((fifo->end[i] == fifo->start + 1) ||
				(fifo->end[i] == 0 && fifo->start == fifo->size - 1))
		   		fifo->rd_overrun_bit |= (1<<i);
		}
		
	}

#ifdef CONFIG_CAN_DEBUG
	if(overrun_bit != fifo->rd_overrun_bit)
		printk(KERN_WARNING "CAN Rx FIFO is full bit = 0x%x\n", fifo->rd_overrun_bit);
#endif
	if(fifo->rd_overrun_bit == fifo->user_counter) {
		printk(KERN_WARNING "CAN Rx FIFO is full bit = 0x%x\n", fifo->rd_overrun_bit);
		retval =  -ENOSPC;
	}
	return retval;
}

static int  can_fifo_empty(can_fifo_t *fifo, int id)
{
	if (fifo->end[id] == -1){
		printk(KERN_ERR "%s: Wicked Error Client number %d \n",__func__,id);
		return -EINVAL;
	}
	return ((fifo->start == fifo->end[id]) ? (-ENODATA) : 0 );
}

/* FIFO initialization, and end[i] = -1 means that this slot is idle. */
int can_fifo_init(can_fifo_t *fifo, int size)
{
	int i;

	fifo->data = (can_frame_t *) kmalloc(size * sizeof(struct can_frame), GFP_KERNEL);

	if (!fifo->data) {
		return -ENOMEM;
	}
	
	fifo->size = size;
	fifo->start = 0;	
	for (i=0; i<CAN_MAX_USER; i++) {
		fifo->end[i] = -1;
	}
	
	spin_lock_init(&fifo->lock);

	return 0;
}

int can_fifo_adduser(can_fifo_t *fifo) 
{
	int i;
	for (i=0; i<CAN_MAX_USER; i++) {
		if (fifo->end[i] == -1) {
			fifo->end[i] = fifo->start;
			fifo->user_counter |= (1<<i);
			return i;
		}
	}
	return -EBUSY;
}

int can_fifo_deluser(can_fifo_t *fifo, int id) 
{

	fifo->end[id] = -1;
	fifo->user_counter &= ~(1<<id);
	fifo->rd_overrun_bit &= ~(1<<id);
	return 0;
}

/* Push a CAN frame into FIFO */
int can_fifo_push(can_fifo_t *fifo, can_frame_t *frame) 
{
   	int retval = 0;

	retval = can_fifo_full_check(fifo);

	memcpy(&fifo->data[fifo->start], frame, sizeof(struct can_frame));

	if (++fifo->start == fifo->size)
		fifo->start = 0;
	return retval;
}

/* ------------------------------------------------------------------------ */
/* POP a CAN frame from FIFO */
can_frame_t * can_fifo_pop(can_fifo_t *fifo, int id, int  *retval) 
{
	int val = can_fifo_empty(fifo, id);
	can_frame_t * frame = NULL;

	*retval =  val;
	
	if (val == 0) {
		frame = &fifo->data[fifo->end[id]];

		/*There is a implicit expression, and wrap around*/
		if (++fifo->end[id] ==  fifo->size)
			fifo->end[id] = 0;
	}
	
	return frame;
}

/* get a CAN free frame pointer from FIFO */
can_frame_t * can_fifo_get(can_fifo_t *fifo, int *retval) 
{
	can_frame_t * frame = NULL;
	
	if((fifo->end[0] == fifo->start + 1) ||(fifo->end[0] == 0 && fifo->start == fifo->size - 1))
		   	*retval =  -EBUSY;
	else
			*retval =  0;
	
	frame = &(fifo->data[fifo->start]);
	if (++fifo->start == fifo->size)
		fifo->start = 0;
	return frame;
}

EXPORT_SYMBOL(can_fifo_push);
EXPORT_SYMBOL(can_fifo_pop); 

