/*
 * drivers/i2c/chips/max6900.c
 *
 * I2C client/driver for the Maxim/Dallas MAX6900 Real-Time Clock chip.
 *
 * Author: Carlos Sanchez <csanchez@mvista.com>
 * Based on M41T00 driver written by Mark A. Greer <mgreer@mvista.com>
 *
 * 2005 (c) MontaVista Software, Inc. This file is licensed under
 * the terms of the GNU General Public License version 2. This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */
/*
 * This i2c client/driver wedges between the drivers/char/genrtc.c RTC
 * interface and the SMBus interface of the i2c subsystem.
 * It would be more efficient to use i2c msgs/i2c_transfer directly but, as
 * recommened in .../Documentation/i2c/writing-clients section
 * "Sending and receiving", using SMBus level communication is preferred.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/i2c.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <linux/workqueue.h>

#include <asm/time.h>
#include <asm/rtc.h>

#define	MAX6900_DRV_NAME		"max6900"
#define	I2C_DF_NOTIFY 0x01

static DECLARE_MUTEX(max6900_mutex);

static struct i2c_driver max6900_driver;
static struct i2c_client *save_client;

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { 0x50, I2C_CLIENT_END };

static struct work_struct set_rtc_time_task;

static struct i2c_client_address_data addr_data = {
	.normal_i2c		= normal_addr,
	.probe			= ignore,
	.ignore			= ignore,
 	.forces			= 0,
};

ulong
max6900_get_rtc_time(void)
{
	s32	sec, min, hour, day, mon, year;
	s32	sec1, min1, hour1, day1, mon1, year1;
	ulong	limit = 10;

	sec = min = hour = day = mon = year = 0;
	sec1 = min1 = hour1 = day1 = mon1 = year1 = 0;

	down(&max6900_mutex);
	do {
		if (((sec = i2c_smbus_read_byte_data(save_client, 0x81)) >= 0)
			&& ((min = i2c_smbus_read_byte_data(save_client, 0x83))
				>= 0)
			&& ((hour = i2c_smbus_read_byte_data(save_client, 0x85))
				>= 0)
			&& ((day = i2c_smbus_read_byte_data(save_client, 0x87))
				>= 0)
			&& ((mon = i2c_smbus_read_byte_data(save_client, 0x89))
				>= 0)
			&& ((year = i2c_smbus_read_byte_data(save_client, 0x8d))
				>= 0)
			&& ((sec == sec1) && (min == min1) && (hour == hour1)
				&& (day == day1) && (mon == mon1)
				&& (year == year1)))

				break;

		sec1 = sec;
		min1 = min;
		hour1 = hour;
		day1 = day;
		mon1 = mon;
		year1 = year;
	} while (--limit > 0);
	up(&max6900_mutex);

	if (limit == 0) {
		dev_warn(&save_client->dev,
			"max6900: can't read rtc chip\n");
		sec = min = hour = day = mon = year = 0;
	}

	sec &= 0x7f;
	min &= 0x7f;
	hour &= 0x3f;
	day &= 0x3f;
	mon &= 0x1f;
	year &= 0xff;

	BCD_TO_BIN(sec);
	BCD_TO_BIN(min);
	BCD_TO_BIN(hour);
	BCD_TO_BIN(day);
	BCD_TO_BIN(mon);
	BCD_TO_BIN(year);

	year += 1900;
	if (year < 1970)
		year += 100;



	return mktime(year, mon, day, hour, min, sec);
}

static void
max6900_set_tlet(ulong arg)
{
	struct rtc_time	tm;
	ulong	nowtime = *(ulong *)arg;
	s32	century;

	to_tm(nowtime, &tm);
	century = tm.tm_year / 100;
	tm.tm_year = tm.tm_year % 100;
	tm.tm_wday++;

	BIN_TO_BCD(tm.tm_sec);
	BIN_TO_BCD(tm.tm_min);
	BIN_TO_BCD(tm.tm_hour);
	BIN_TO_BCD(tm.tm_mon);
	BIN_TO_BCD(tm.tm_mday);
	BIN_TO_BCD(tm.tm_year);
	BIN_TO_BCD(tm.tm_wday);
	BIN_TO_BCD(century);

	down(&max6900_mutex);
	if ((i2c_smbus_write_byte_data(save_client, 0x80, tm.tm_sec & 0x7f) < 0)
		|| (i2c_smbus_write_byte_data(save_client, 0x82, tm.tm_min & 0x7f)
			< 0)
		|| (i2c_smbus_write_byte_data(save_client, 0x84, tm.tm_hour & 0x3f)
			< 0)
		|| (i2c_smbus_write_byte_data(save_client, 0x86, tm.tm_mday & 0x3f)
			< 0)
		|| (i2c_smbus_write_byte_data(save_client, 0x88, tm.tm_mon & 0x1f)
			< 0)
		|| (i2c_smbus_write_byte_data(save_client, 0x8A, tm.tm_wday & 0x0f)
			< 0)
		|| (i2c_smbus_write_byte_data(save_client, 0x8C, tm.tm_year & 0xff)
			< 0)
		|| (i2c_smbus_write_byte_data(save_client, 0x92, century & 0xff)
			< 0))

		dev_warn(&save_client->dev,"max6900: can't write to rtc chip\n");

	up(&max6900_mutex);
	return;
}

ulong	new_time;

int
max6900_set_rtc_time(ulong nowtime)
{
	new_time = nowtime;

	if (in_interrupt())
		schedule_work(&set_rtc_time_task);
	else
		max6900_set_tlet((ulong)&new_time);

	return 0;
}

/*
 *****************************************************************************
 *
 *	Driver Interface
 *
 *****************************************************************************
 */
static int
max6900_probe(struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *client;
	int rc;

	client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (!client)
		return -ENOMEM;

	memset(client, 0, sizeof(struct i2c_client));
	strncpy(client->name, MAX6900_DRV_NAME, I2C_NAME_SIZE);
/* 	client->id = max6900_driver.id; */
	client->flags = I2C_DF_NOTIFY;
	client->addr = addr;
	client->adapter = adap;
	client->driver = &max6900_driver;

	if ((rc = i2c_attach_client(client)) != 0) {
		kfree(client);
		return rc;
	}

	INIT_WORK(&set_rtc_time_task,
			(void (*)(void *))&max6900_set_tlet);

	save_client = client;
	return 0;
}

static int
max6900_attach(struct i2c_adapter *adap)
{
	int	rc;
	rc = i2c_probe(adap, &addr_data, max6900_probe);
	return rc;
}

static int
max6900_detach(struct i2c_client *client)
{
	int	rc;

	if ((rc = i2c_detach_client(client)) == 0) {
		kfree(i2c_get_clientdata(client));
		flush_scheduled_work();

	}
	return rc;
}

static struct i2c_driver max6900_driver = {
	.id		= I2C_DRIVERID_MAX6900,
	.attach_adapter	= max6900_attach,
	.detach_client	= max6900_detach,
};

static int __init
max6900_init(void)
{
	return i2c_add_driver(&max6900_driver);
}

static void __exit
max6900_exit(void)
{
	i2c_del_driver(&max6900_driver);
	return;
}

module_init(max6900_init);
module_exit(max6900_exit);

MODULE_AUTHOR("Carlos Sanchez <csanchez@mvista.com>");
MODULE_DESCRIPTION("MAX6900 RTC I2C Client Driver");
MODULE_LICENSE("GPL");
