/*
 * Copyright (C) 2007, Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * Based on swarm.c:
 *
 *  Copyright (C) 2001, 2002, 2003 Broadcom Corporation
 *  Copyright (C) 2004 MontaVista Software Inc.
 *       Author:	Manish Lachwani, mlachwani@mvista.com
 *  Copyright (C) 2004  MIPS Technologies, Inc.  All rights reserved.
 *       Author: Maciej W. Rozycki <macro@mips.com>
 *  Copyright (c) 2006  Maciej W. Rozycki
 *
 * swarm.c was derrived loosely from ide-pmac.c:
 *
 *  Copyright (C) 1998 Paul Mackerras.
 *  Copyright (C) 1995-1998 Mark Lord
 *
 * Most of the IDE setup code (io_ports, ioaddr etc.) comes from a support
 * patch to 8349e_itx.c, provided by Freescale.
 */

/*
 * Driver intended for and tested on MPC 8349e mitx's compact flash card,
 * though it's likely to work for other IDE chips that are represented
 * in an open firmware device node.
 */

#include <linux/module.h>
#include <linux/ide.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/platform_device.h>

#include <asm/io.h>

#define MPC83xx_IDE_ALT_BASE_OFFSET	0x200
#define MPC83xx_IDE_CONTROL_OFFSET	0xc
#define MPC83xx_IDE_IRQ_OFFSET		0xe

#define DRV_NAME "ide-fsl"

MODULE_LICENSE("GPL");

static char fsl_ide_string[] = DRV_NAME;
static struct platform_device *fsl_ide_dev;

/*
 * fsl_ide_probe: Check openfirmware tree for the IDE device and map it if found
 */
static int __devinit fsl_ide_probe(struct device *dev)
{
	struct device_node *np;
	struct resource res;
	ide_hwif_t *hwif;
	unsigned long ioaddr;
	unsigned int irq;
	int i;


	/* Find an empty slot.  */
	for (i = 0; i < MAX_HWIFS; i++)
		if (!ide_hwifs[i].io_ports[IDE_DATA_OFFSET])
			break;
	if (i >= MAX_HWIFS) {
		printk(KERN_ERR DRV_NAME ": no free slot for interface\n");
		return -ENOMEM;
	}

	hwif = ide_hwifs + i;

	/* Get device info from OF tree */
	np = of_find_compatible_node(NULL, "ide", "fsl-ide");
	if (!np)
		return -ENODEV;

	irq = irq_of_parse_and_map(np, 0);

	if (of_address_to_resource(np, 0, &res)) {
		of_node_put(np);
		return -ENODEV;
	}

	of_node_put(np);

	/* Initialize the device then call generic IDE code */
	hwif->irq = irq;
	hwif->io_ports[IDE_DATA_OFFSET] = (unsigned long) ioremap(res.start, 2);

	default_hwif_mmiops(hwif);

	hwif->mmio = 1;
	hwif->noprobe = 0;

	/* Disable probing for slave, otherwise the device probing routines
	 * might freeze for 35 seconds waiting for a non-existing deivice */
	hwif->drives[1].noprobe = 1;

	ioaddr = hwif->io_ports[IDE_DATA_OFFSET]+2;
	for (i = IDE_ERROR_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hwif->io_ports[i] = ioaddr;
		ioaddr += 2;
	}

	hwif->io_ports[IDE_CONTROL_OFFSET] = hwif->io_ports[IDE_DATA_OFFSET]+MPC83xx_IDE_ALT_BASE_OFFSET+MPC83xx_IDE_CONTROL_OFFSET;
	hwif->io_ports[IDE_IRQ_OFFSET] = hwif->io_ports[IDE_DATA_OFFSET]+MPC83xx_IDE_ALT_BASE_OFFSET+MPC83xx_IDE_IRQ_OFFSET;

	memcpy(hwif->hw.io_ports, hwif->io_ports, sizeof(hwif->io_ports));
	hwif->hw.irq = hwif->irq;

	probe_hwif_init(hwif);
	dev_set_drvdata(dev, hwif);

	return 0;
}

static int fsl_ide_remove (struct device * dev) {
	ide_hwif_t *hwif = dev_get_drvdata(dev);

	ide_unregister(ide_hwifs - hwif);
	iounmap(hwif->io_ports[IDE_DATA_OFFSET]);

	return 0;
}

static struct device_driver fsl_ide_driver = {
	.name	= fsl_ide_string,
	.bus	= &platform_bus_type,
	.probe	= fsl_ide_probe,
	.remove = fsl_ide_remove,
};

static void fsl_ide_platform_release(struct device *device)
{
	struct platform_device *pldev;

	/* free device */
	pldev = to_platform_device(device);
	kfree(pldev);
}

static int __devinit fsl_ide_init_module(void)
{
	struct platform_device *pldev;
	int err;

	printk(KERN_INFO "FSL Open Firmware IDE driver\n");

	if (driver_register(&fsl_ide_driver)) {
		printk(KERN_ERR "Driver registration failed\n");
		err = -ENODEV;
		goto out;
	}

        if (!(pldev = kmalloc(sizeof (*pldev), GFP_KERNEL))) {
		err = -ENOMEM;
		goto out_unregister_driver;
	}

	memset (pldev, 0, sizeof (*pldev));
	pldev->name		= fsl_ide_string;
	pldev->id		= 0;
	pldev->dev.release	= fsl_ide_platform_release;

	if (platform_device_register(pldev)) {
		err = -ENODEV;
		goto out_free_pldev;
	}

        if (!pldev->dev.driver) {
		/*
		 * The driver was not bound to this device, there was
                 * no hardware at this address. Unregister it, as the
		 * release fuction will take care of freeing the
		 * allocated structure
		 */
		platform_device_unregister (pldev);
	}

	fsl_ide_dev = pldev;

	return 0;

out_free_pldev:
	kfree(pldev);

out_unregister_driver:
	driver_unregister(&fsl_ide_driver);
out:
	return err;
}

static void fsl_ide_exit_module(void)
{
	driver_unregister(&fsl_ide_driver);
	platform_device_unregister(fsl_ide_dev);
}

module_init(fsl_ide_init_module);
module_exit(fsl_ide_exit_module);
