/*
 * Copyright 2007 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * PATA driver for mpc512x
 *
 * Based on:
 *     linux/drivers/ide/arm/mxc_ide.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*!
 * @file mcp512x_ide.c
 *
 * @brief ATA driver
 *
 * @ingroup ATA
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/ide.h>
#include <linux/init.h>
#include <linux/ide.h>
#include <linux/clk.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/delay.h>

#include <asm/of_device.h>
#include <asm/of_platform.h>
#include "mpc512x.h"

#define DRV_NAME	"mpc512x_ide"

extern void gpio_ata_inactive(void);

/* List of registered interfaces */
static ide_hwif_t *ifs[1];

/*
 * This structure contains the timing parameters for
 * ATA bus timing in the 5 PIO modes.  The timings
 * are in nanoseconds, and are converted to clock
 * cycles before being stored in the ATA controller
 * timing registers.
 */
static struct {
	short t0, t1, t2_8, t2_16, t2i, t4, t9, tA;
} pio_specs[] = {
	[0] = {
		.t0 = 600, .t1 = 70,
		.t2_8 = 290, .t2_16 = 165, .t2i = 0,
		.t4 = 30, .t9 = 20, .tA = 50
	},
	[1] = {
		.t0 = 383, .t1 = 50,
		.t2_8 = 290, .t2_16 = 125, .t2i = 0,
		.t4 = 20, .t9 = 15, .tA = 50
	},
	[2] = {
		.t0 = 240, .t1 = 30,
		.t2_8 = 290, .t2_16 = 100, .t2i = 0,
		.t4 = 15, .t9 = 10, .tA = 50
	},
	[3] = {
		.t0 = 180, .t1 = 30,
		.t2_8 = 80, .t2_16 = 80, .t2i = 0,
		.t4 = 10, .t9 = 10, .tA = 50
	},
	[4] = {
		.t0 = 120, .t1 = 25,
		.t2_8 = 70, .t2_16 = 70, .t2i = 0,
		.t4 = 10, .t9 = 10, .tA = 50
	}
};

#define NR_PIO_SPECS (sizeof pio_specs / sizeof pio_specs[0])

/*!
 * Calculate values for the ATA bus timing registers and store
 * them into the hardware.
 *
 * @param       mode        Selects PIO, MDMA or UDMA modes
 *
 * @param       speed       Specifies the sub-mode number
 *
 * @return      EINVAL      speed out of range, or illegal mode
 */
static int set_ata_bus_timing(struct fsl_ata_priv *priv, int speed,
			      enum ata_mode mode)
{
	/* get the bus clock cycle time, in ns */
	int T = 1 * 1000 * 1000 * 1000 / clk_get_rate(priv->ata_clk);
	union fsl_ata_time_cfg cfg0, cfg1, cfg2;
	/* every mode gets the same t_off and t_on */

	GET_TIME_CFG(&cfg0, FSL_ATA_TIME_OFF);
	cfg0.bytes.field1 = 3;
	cfg0.bytes.field2 = 3;
	SET_TIME_CFG(&cfg0, 3, FSL_ATA_TIME_OFF);

	switch (mode) {
	case PIO:
		if (speed < 0 || speed >= NR_PIO_SPECS) {
			return -EINVAL;
		}
		cfg0.bytes.field3 = (pio_specs[speed].t1 + T) / T;
		cfg0.bytes.field4 = (pio_specs[speed].t2_8 + T) / T;

		cfg1.bytes.field1 = (pio_specs[speed].t2_8 + T) / T;
		cfg1.bytes.field2 = (pio_specs[speed].tA + T) / T + 2;
		cfg1.bytes.field3 = 1;
		cfg1.bytes.field4 = (pio_specs[speed].t4 + T) / T;

		GET_TIME_CFG(&cfg2, FSL_ATA_TIME_9);
		cfg2.bytes.field1 = (pio_specs[speed].t9 + T) / T;

		SET_TIME_CFG(&cfg0, 0x0C, FSL_ATA_TIME_OFF);
		SET_TIME_CFG(&cfg1, 0x0F, FSL_ATA_TIME_2r);
		SET_TIME_CFG(&cfg2, 0x01, FSL_ATA_TIME_9);
		break;
	case MDMA:
	case UDMA:
	default:
		return -EINVAL;
	}
	return 0;
}

/*!
 * Placeholder for code to make any hardware tweaks
 * necessary to select a drive.  Currently we are
 * not aware of any.
 */
static void fsl_ata_selectproc(ide_drive_t *drive)
{
	return;
}

/*!
 * Called to set the PIO mode.
 *
 * @param   drive       Specifies the drive
 * @param   pio    Specifies the PIO mode number desired
 */
static void fsl_ata_tune(ide_drive_t *drive, u8 pio)
{
	struct fsl_ata_priv *priv =
	    (struct fsl_ata_priv *)HWIF(drive)->hwif_data;

	set_ata_bus_timing(priv, pio, PIO);
}

/*!
 * Hardware-specific interrupt service routine for the ATA driver,
 * called mainly just to dismiss the interrupt at the hardware, and
 * to indicate whether there actually was an interrupt pending.
 *
 * The generic IDE related interrupt handling is done by the IDE layer.
 */
static int fsl_ata_ack_intr(struct hwif_s *hw)
{
	struct fsl_ata_priv *priv = hw->hwif_data;
	unsigned char status = ATA_RAW_READ(FSL_ATA_INTR_PENDING);
	unsigned char enable = ATA_RAW_READ(FSL_ATA_INTR_ENABLE);

	/*
	 * The only interrupts we can actually dismiss are the FIFO conditions.
	 * INTRQ comes from the bus, and must be dismissed according to IDE
	 * protocol, which will be done by the IDE layer, even when DMA
	 * is invovled (DMA can see it, but can't dismiss it).
	 */
	ATA_RAW_WRITE(status, FSL_ATA_INTR_CLEAR);

	if (status & enable & ~FSL_ATA_INTR_ATA_INTRQ2) {
		printk(KERN_ERR "mpc512x_ide_ack_intr: unexpected interrupt, "
		       "status=0x%02X\n", status);
	}

	return status ? 1 : 0;
}

static void fsl_ata_resetproc(struct fsl_ata_priv *priv)
{
	ATA_RAW_WRITE(0x00, FSL_ATA_CONTROL);
	udelay(100);
	ATA_RAW_WRITE(FSL_ATA_CTRL_ATA_RST_B, FSL_ATA_CONTROL);
	udelay(100);
}

/*!
 * Directs the IDE INTRQ signal to DMA or to the CPU
 *
 * @param       hwif       Specifies the IDE controller
 *
 * @param       which       \b INTRQ_DMA or \b INTRQ_MCU
 *
 */
static void fsl_ata_set_intrq(ide_hwif_t *hwif, int which)
{
	struct fsl_ata_priv *priv = (struct fsl_ata_priv *) hwif->hwif_data;
	unsigned char enable = ATA_RAW_READ(FSL_ATA_INTR_ENABLE);

	switch (which) {
	case INTRQ_DMA:
		enable &= ~FSL_ATA_INTR_ATA_INTRQ2;
		enable |= FSL_ATA_INTR_ATA_INTRQ1;
		break;
	case INTRQ_MCU:
		enable &= ~FSL_ATA_INTR_ATA_INTRQ1;
		enable |= FSL_ATA_INTR_ATA_INTRQ2;
		break;
	}
	priv->enable = enable;

	ATA_RAW_WRITE(enable, FSL_ATA_INTR_ENABLE);
}

/*!
 * Masks drive interrupts temporarily at the hardware level
 *
 * @param       drive       Specifies the drive
 *
 * @param       mask        1 = disable interrupts, 0 = re-enable
 *
 */
static void fsl_ata_maskproc(ide_drive_t *drive, int mask)
{
	struct fsl_ata_priv *priv =
	    (struct fsl_ata_priv *) (HWIF(drive)->hwif_data);
	BUG_ON(!priv);

	if (mask) {
		ATA_RAW_WRITE(0, FSL_ATA_INTR_ENABLE);
	} else {
		ATA_RAW_WRITE(priv->enable, FSL_ATA_INTR_ENABLE);
	}
}

/*!
 * MPC512X-specific IDE registration helper routine.  Among other things,
 * we tell the IDE layer where our standard IDE drive registers are,
 * through the \b io_ports array.  The IDE layer sends commands to and
 * reads status directly from attached IDE drives through these drive
 * registers.
 *
 * @param   base   Base address of memory mapped IDE standard drive registers
 *
 * @param   aux    Address of the auxilliary ATA control register
 *
 * @param   irq    IRQ number for our hardware
 *
 * @param   hwifp  Pointer to hwif structure to be updated by the IDE layer
 *
 * @param   priv   Pointer to private structure
 *
 * @return  ENODEV if no drives are present, 0 otherwise
 */
static int __init
fsl_ata_register(unsigned long base, unsigned long aux, int irq,
		 ide_hwif_t **hwifp, struct fsl_ata_priv *priv)
{
	int i = 0;
	hw_regs_t hw;
	ide_hwif_t *hwif = &ide_hwifs[0];
	const int regsize = 4;

	memset(&hw, 0, sizeof(hw));

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw.io_ports[i] = (unsigned long)base;
		base += regsize;
	}
	hw.io_ports[IDE_CONTROL_OFFSET] = aux;
	hw.irq = irq;
	hw.ack_intr = &fsl_ata_ack_intr;

	*hwifp = hwif;
	ide_register_hw(&hw, hwifp);

	hwif->selectproc = &fsl_ata_selectproc;
	hwif->tuneproc = &fsl_ata_tune;
	hwif->maskproc = &fsl_ata_maskproc;
	hwif->hwif_data = (void *)priv;
	fsl_ata_set_intrq(hwif, INTRQ_MCU);

	return 0;
}

/*!
 * Driver initialization routine.  Prepares the hardware for ATA activity.
 */
static int __init mpc512x_ata_probe(struct of_device *op,
				    const struct of_device_id *match)
{
	int index = 0;
	struct fsl_ata_priv *priv;
	struct resource res_mem;
	int rv;

	/*
	 * Allocate a private structure
	 */
	priv = devm_kzalloc(&op->dev, sizeof(*priv), GFP_KERNEL);
	if (priv == NULL) {
		printk(KERN_ERR DRV_NAME ": "
		       "Error while allocating private structure\n");
		return ENOMEM;
	}

	priv->dev = NULL;
	priv->dma_read_chan = -1;
	priv->dma_write_chan = -1;

	/* Get IRQ and register */
	rv = of_address_to_resource(op->node, 0, &res_mem);
	if (rv) {
		printk(KERN_ERR DRV_NAME ": "
		       "Error while parsing device node resource\n");
		return rv;
	}

	priv->ata_irq = irq_of_parse_and_map(op->node, 0);
	if (priv->ata_irq == NO_IRQ) {
		printk(KERN_ERR DRV_NAME ": " "Error while mapping the irq\n");
		return -EINVAL;
	}

	/* Request mem region */
	if (!devm_request_mem_region(&op->dev, res_mem.start,
				     1 + res_mem.end - res_mem.start,
				     DRV_NAME)) {
		printk(KERN_ERR DRV_NAME ": "
		       "Error while requesting mem region\n");
		rv = -EBUSY;
		goto err;
	}

	/* mapin regs */
	priv->ata_regs = devm_ioremap(&op->dev, res_mem.start,
				      1 + res_mem.end - res_mem.start);
	if (!priv->ata_regs) {
		printk(KERN_ERR DRV_NAME ": "
		       "Error while mapping register set\n");
		rv = -ENOMEM;
		goto err;
	}

	priv->ata_clk = clk_get(&op->dev, "ata_clk");
	clk_enable(priv->ata_clk);

	/* Deassert the reset bit to enable the interface */
	ATA_RAW_WRITE(FSL_ATA_CTRL_ATA_RST_B, FSL_ATA_CONTROL);

	/* Set initial timing and mode */
	set_ata_bus_timing(priv, 0, PIO);

	/* Reset the interface */
	fsl_ata_resetproc(priv);

	/*
	 * Enable hardware interrupts.
	 * INTRQ2 goes to us, so we enable it here, but we'll need to ignore
	 * it when DMA is doing the transfer.
	 */
	ATA_RAW_WRITE(FSL_ATA_INTR_ATA_INTRQ2, FSL_ATA_INTR_ENABLE);

	/*
	 * Now register
	 */
	index =
	    fsl_ata_register((unsigned long)FSL_ATA_IO,
			     (unsigned long)FSL_ATA_CTL, priv->ata_irq, &ifs[0],
			     priv);
	if (index == -1) {
		printk(KERN_ERR "Unable to register the MPC512X IDE driver\n");
		ATA_RAW_WRITE(0, FSL_ATA_INTR_ENABLE);
		kfree(priv);
		return ENODEV;
	}
#ifdef ATA_USE_IORDY
	/* turn on IORDY protocol */

	udelay(25);
	ATA_RAW_WRITE(FSL_ATA_CTRL_ATA_RST_B | FSL_ATA_CTRL_IORDY_EN,
		      FSL_ATA_CONTROL);
#endif

	return 0;

	/* error exit */
err:
	irq_dispose_mapping(priv->ata_irq);
	return rv;
}

/*!
 * Driver exit routine.  Clean up.
 */
static int __exit mpc512x_ata_remove(struct of_device *op)
{
	ide_hwif_t *hwif = ifs[0];
	struct fsl_ata_priv *priv;

	BUG_ON(!hwif);
	priv = (struct fsl_ata_priv *) hwif->hwif_data;
	BUG_ON(!priv);

	/*
	 * Unregister the interface at the IDE layer.  This should shut
	 * down any drives and pending I/O.
	 */
	ide_unregister(hwif->index);

	/*
	 * Disable hardware interrupts.
	 */
	ATA_RAW_WRITE(0, FSL_ATA_INTR_ENABLE);

	/*
	 * Turn off the clock
	 */
	clk_disable(priv->ata_clk);
	clk_put(priv->ata_clk);

	/*
	 * Cleanup the irq_mapping
	 */
	irq_dispose_mapping(priv->ata_irq);

	/*
	 * Free the private structure.
	 */
	kfree(priv);
	hwif->hwif_data = NULL;

	return 0;
}

#ifdef CONFIG_PM

static int mpc512x_ata_suspend(struct of_device *op, pm_message_t state)
{
	return 0;		/* FIXME */
}

static int mpc512x_ata_resume(struct of_device *op)
{
	return 0;		/* FIXME */
}

#else
#define mpc512x_ata_suspend NULL
#define mpc512x_ata_resume NULL
#endif

static struct of_device_id mpc512x_ata_of_match[] = {
	{
	 .compatible = "mpc512x-pata",
	 },
	{},
};

static struct of_platform_driver mpc512x_ata_of_platform_driver = {
	.owner = THIS_MODULE,
	.name = DRV_NAME,
	.match_table = mpc512x_ata_of_match,
	.probe = mpc512x_ata_probe,
	.remove = mpc512x_ata_remove,
	.suspend = mpc512x_ata_suspend,
	.resume = mpc512x_ata_resume,
	.driver = {
		   .name = DRV_NAME,
		   .owner = THIS_MODULE,
		   },
};

static int __init fsl_ata_init(void)
{
	printk(KERN_INFO
	       "MPC512X: IDE driver, (c) 2004-2007 Freescale Semiconductor\n");
	return of_register_platform_driver(&mpc512x_ata_of_platform_driver);
}

static void __exit fsl_ata_exit(void)
{
	of_unregister_platform_driver(&mpc512x_ata_of_platform_driver);
}

module_init(fsl_ata_init);
module_exit(fsl_ata_exit);

MODULE_AUTHOR("Freescale Semiconductor, Inc.");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION
    ("Freescale MPC512X IDE "
     "(based on Simtec BAST IDE driver by Ben Dooks <ben@simtec.co.uk>)");
