/*
 * drivers/ide/ppc/mpc52xx_ide.h
 *
 * Driver for the Freescale MPC52xx on-chip IDE interface
 *
 *
 * Copyright (C) 2006 Sylvain Munaut <tnt@246tNt.com>
 * Copyright (C) 2003 Mipsys - Benjamin Herrenschmidt
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2. This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */

#undef REALLY_SLOW_IO	/* most systems can safely undef this */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/ide.h>
#include <linux/notifier.h>
#include <linux/reboot.h>   
#include <linux/pci.h>
#include <linux/adb.h>
#include <linux/pmu.h>
#include <linux/scatterlist.h>

#include <asm/prom.h>
#include <asm/io.h>
#include <asm/dbdma.h>
#include <asm/ide.h>
#include <asm/pci-bridge.h>
#include <asm/machdep.h> 
#include <asm/pmac_feature.h>
#include <asm/sections.h>
#include <asm/irq.h>

#include <asm/ppcboot.h>
#include <asm/mpc52xx.h>

#include "mpc52xx_ide.h"

#define CONFIG_BLK_DEV_IDE_MPC52xx_SHOWDETAILS 1
#define MPC52xx_GPIO_OFFSET	0x0b00
#define MPC52xx_GPIO_SIZE	0x040
#define MPC52xx_MBAR		0xf0000000      /* Phys address */
#define MPC52xx_PA(x)	 	((phys_addr_t)(MPC52xx_MBAR + (x)))

/* Private structures used by the driver */
struct mpc52xx_ata_timings {
	u32 pio1;
	u32 pio2;
	u32 mdma1;
	u32 mdma2;
	u32 udma1;
	u32 udma2;
	u32 udma3;
	u32 udma4;
	u32 udma5;
	int using_udma;
};

struct mpc52xx_ide_priv {
	unsigned int ipb_period;	/* in ps */
	struct mpc52xx_ata __iomem *ata_regs;
	struct mpc52xx_ata_timings  timings[2];
};

/* ATAPI-4 PIO specs (arranged for the 5200, cfr User Manual) */
/* numbers in ns, extrapolation done by code */
static int ataspec_t0[5]    = {600, 383, 240, 180, 120};
static int ataspec_t1[5]    = { 70,  50,  30,  30,  25};
static int ataspec_t2_8[5]  = {290, 290, 290,  80,  70};
static int ataspec_t2_16[5] = {165, 125, 100,  80,  70};
static int ataspec_t2i[5]   = {  0,   0,   0,  70,  25};
static int ataspec_t4[5]    = { 30,  20,  15,  10,  10};
static int ataspec_ta[5]    = { 35,  35,  35,  35,  35};

/* Helpers to compute timing parameters */
#define CALC_CLK_VALUE_UP(c,v) (((v) + c - 1) / c)
#define DRV_NAME "ide-mpc52xx"
static char mpc52xx_ide_string[] = DRV_NAME;
static struct platform_device *mpc52xx_ide_dev;


/* ======================================================================== */
/* IDE Driver & Aux functions                                               */
/* ======================================================================== */

static void
mpc52xx_ide_apply_timing(
	struct mpc52xx_ata __iomem *regs, struct mpc52xx_ata_timings *timing)
{
	out_be32(&regs->pio1,  timing->pio1);
	out_be32(&regs->pio2,  timing->pio2);
	out_be32(&regs->mdma1, timing->mdma1);
	out_be32(&regs->mdma2, timing->mdma2);
	out_be32(&regs->udma1, timing->udma1);
	out_be32(&regs->udma2, timing->udma2);
	out_be32(&regs->udma3, timing->udma3);
	out_be32(&regs->udma4, timing->udma4);
	out_be32(&regs->udma5, timing->udma5);
}

static void
mpc52xx_ide_compute_pio_timing(
	struct mpc52xx_ata_timings *timing, unsigned int ipb_period, u8 pio)
{
	u32 t0, t2_8, t2_16, t2i, t4, t1, ta;

		/* We add 1 as a 'margin' */
	t0    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t0[pio]);
	t2_8  = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t2_8[pio]);
	t2_16 = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t2_16[pio]);
	t2i   = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t2i[pio]);
	t4    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t4[pio]);
	t1    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t1[pio]);
	ta    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_ta[pio]);

	timing->pio1 = (t0 << 24) | (t2_8 << 16) | (t2_16 << 8) | (t2i);
	timing->pio2 = (t4 << 24) | (t1 << 16) | (ta << 8);
}


static void
mpc52xx_ide_tuneproc(ide_drive_t *drive, u8 pio)
{
	struct mpc52xx_ide_priv *priv = drive->hwif->hwif_data;
	struct mpc52xx_ata __iomem *regs = priv->ata_regs;
	int w = drive->select.b.unit & 0x01;

	pio = ide_get_best_pio_mode(drive, pio, 5, NULL);

	printk("%s: Setting PIO %d timings\n", drive->name, pio);

	mpc52xx_ide_compute_pio_timing(&priv->timings[w], priv->ipb_period, pio);

	if (drive->select.all == HWIF(drive)->INB(IDE_SELECT_REG))
		mpc52xx_ide_apply_timing(regs, &priv->timings[w]);

		/* Should we do it here or only in speedproc ? */
	ide_config_drive_speed(drive, pio + XFER_PIO_0);
}

static int
mpc52xx_ide_speedproc(ide_drive_t *drive, u8 speed)
{
	/* Configure PIO Mode */
	if (speed >= XFER_PIO_0 && speed <= XFER_PIO_4) {
		mpc52xx_ide_tuneproc(drive, speed - XFER_PIO_0);
		return 0;
	}

	/* DMA settings currently unsupported */
	printk(KERN_ERR
		"mpc52xx-ide: speedproc called with unsupported mode %d\n",
		speed);

	return 1;
}

static void
mpc52xx_ide_selectproc(ide_drive_t *drive)
{
	/* Change the PIO timings to the ones of the
	   currently selected drive */
	struct mpc52xx_ide_priv *priv = drive->hwif->hwif_data;
	struct mpc52xx_ata __iomem *regs = priv->ata_regs;

	mpc52xx_ide_apply_timing(regs,
		&priv->timings[drive->select.b.unit & 0x01]);
}


static int
mpc52xx_ide_setup(
	struct mpc52xx_ata __iomem *regs, struct mpc52xx_ide_priv *priv)
{
	int tslot;
	unsigned long ipb_freq = 133 *1000 *1000;
	/* All sample code do this */
	out_be32(&regs->share_cnt, 0);

	/* Configure & Reset host */
	out_be32(&regs->config,
		MPC52xx_ATA_HOSTCONF_IE |
		MPC52xx_ATA_HOSTCONF_IORDY |
		MPC52xx_ATA_HOSTCONF_SMR |
		MPC52xx_ATA_HOSTCONF_FR);
	udelay(10);
	out_be32(&regs->config,
		MPC52xx_ATA_HOSTCONF_IE |
		MPC52xx_ATA_HOSTCONF_IORDY);

	/* Get IPB bus period */
	priv->ipb_period = 1000000000 / (ipb_freq/1000);

	/* Try to set the time slot to around 1us = 1000000 ps */
	tslot = CALC_CLK_VALUE_UP(priv->ipb_period, 1000000);
	out_be32(&regs->share_cnt, tslot << 16);

#ifdef CONFIG_BLK_DEV_IDE_MPC52xx_SHOWDETAILS
	printk("ipb=%dMHz, set clock period to %d ps\n",(int)(ipb_freq/1000000),
			priv->ipb_period);
#endif

	/* Init imings to PIO0 (safest) */
	memset(priv->timings, 0x00, 2*sizeof(struct mpc52xx_ata_timings));

	mpc52xx_ide_compute_pio_timing(&priv->timings[0], priv->ipb_period, 0);
	mpc52xx_ide_compute_pio_timing(&priv->timings[1], priv->ipb_period, 0);

	mpc52xx_ide_apply_timing(regs, &priv->timings[0]);

	return 0;
}

static void
mpc52xx_ide_setup_hwif_ports(hw_regs_t *hw, struct mpc52xx_ata __iomem *regs)
{
	/* It's MMIO and we handle all the io ops ourself, */
	/* so theses address are really virtual addresses  */
	hw->io_ports[IDE_DATA_OFFSET]    = (unsigned long) &regs->tf_data;
	hw->io_ports[IDE_ERROR_OFFSET]   = (unsigned long) &regs->tf_features;
	hw->io_ports[IDE_NSECTOR_OFFSET] = (unsigned long) &regs->tf_sec_count;
	hw->io_ports[IDE_SECTOR_OFFSET]  = (unsigned long) &regs->tf_sec_num;
	hw->io_ports[IDE_LCYL_OFFSET]    = (unsigned long) &regs->tf_cyl_low;
	hw->io_ports[IDE_HCYL_OFFSET]    = (unsigned long) &regs->tf_cyl_high;
	hw->io_ports[IDE_SELECT_OFFSET]  = (unsigned long) &regs->tf_dev_head;
	hw->io_ports[IDE_STATUS_OFFSET]  = (unsigned long) &regs->tf_command;
	hw->io_ports[IDE_CONTROL_OFFSET] = (unsigned long) &regs->tf_control;
}

/* ======================================================================== */
/* Platform Driver                                                          */
/* ======================================================================== */
/*
 * fsl_ide_probe: Check openfirmware tree for the IDE device and map it if found
 */
static int __devinit mpc52xx_ide_probe(struct device *dev)
{
	/* Vars */
	ide_hwif_t *hwif;
	struct device_node *np;
	struct mpc52xx_ide_priv *priv = NULL;
	struct mpc52xx_gpio __iomem *gpio_regs = NULL;
	struct mpc52xx_ata __iomem *ata_regs = NULL;
	int ata_irq;
	struct resource res;
	size_t res_size;
	u32 reg;
	int i, rv;

	/* Get an empty slot */
	for (i=0; i<MAX_HWIFS && ide_hwifs[i].io_ports[IDE_DATA_OFFSET]; i++);
	if (i >= MAX_HWIFS) {
		printk(KERN_ERR "mpc52xx-ide: No free hwif slot !\n");
		return -ENOMEM;
	}

	hwif = &ide_hwifs[i];

	/* Check port muxing for compatibility; there may be some bug by czm*/
	gpio_regs = ioremap(MPC52xx_PA(MPC52xx_GPIO_OFFSET), MPC52xx_GPIO_SIZE);
	if (!gpio_regs) {
		printk(KERN_ERR
			"mpc52xx-ide: Unable to ioremap MPC52xx_GPIO zone\n");
		return -ENOMEM;
	}

	reg = in_be32(&gpio_regs->port_config);
	iounmap(gpio_regs);

	if (!(reg & 0x03000000ul)) {
		printk(KERN_ERR
			"mpc52xx-ide: Invalid port config, no ATA CS !\n");
		return -ENODEV;
	}

	/* Get device info from OF tree */
	np = of_find_compatible_node(NULL, "ata", "mpc5200b-ata");
	if (!np) {
		return -ENODEV;
	}

	ata_irq = irq_of_parse_and_map(np, 0);

	if (of_address_to_resource(np, 0, &res)) {
		of_node_put(np);
		return -ENODEV;
	}

	of_node_put(np);
	res_size = res.end - res.start + 1;

	if (ata_irq < 0) {
		printk(KERN_ERR "mpc52xx-ide: Invalid resource set !\n");
		return -EINVAL;
	}

	if (!request_mem_region(res.start,res_size, "mpc52xx-ide")) {
		printk(KERN_ERR "mpc52xx-ide: Memory zone unavailable !\n");
		return -EBUSY;
	}
	
	ata_regs = ioremap(res.start, res_size);
	if (!ata_regs) {
		printk(KERN_ERR
			"mpc52xx-ide: Unable to ioremap ATA registers\n");
		rv = -ENOMEM;
		goto error;
	}

	/* Setup private structure */
	priv = kmalloc(sizeof(struct mpc52xx_ide_priv), GFP_ATOMIC);
	if (!priv) {
		printk(KERN_ERR
			"mpc52xx-ide: Can't allocate private structure !\n");
		rv = -ENOMEM;
		goto error;
	}

	memset(priv, 0, sizeof *priv);
	priv->ata_regs = ata_regs;

	/* Setup the ATA controller */
	rv = mpc52xx_ide_setup(ata_regs, priv);
	if (rv) {
		printk(KERN_ERR "mpc52xx-ide: Controller setup failed !\n");
		goto error;
	}

	/* Setup the ATA controller */
	rv = mpc52xx_ide_setup(ata_regs, priv);
	if (rv) {
		printk(KERN_ERR "mpc52xx-ide: Controller setup failed !\n");
		goto error;
	}

	/* Setup the hwif structure */
	hwif->irq = ata_irq;
	hwif->mmio = 2;
	mpc52xx_ide_setup_hwif_iops(hwif);
	mpc52xx_ide_setup_hwif_ports(&hwif->hw, ata_regs);
	memcpy(hwif->io_ports, hwif->hw.io_ports, sizeof(hwif->io_ports));

	hwif->atapi_dma = 0;
	hwif->ultra_mask = 0x00;
	hwif->mwdma_mask = 0x00;
	hwif->swdma_mask = 0x00;
	hwif->chipset = ide_mpc52xx;
	hwif->tuneproc = mpc52xx_ide_tuneproc;
	hwif->speedproc = mpc52xx_ide_speedproc;
	hwif->selectproc = mpc52xx_ide_selectproc;
	hwif->noprobe = 0;
	hwif->hold = 1;
	hwif->autodma = 0;
	hwif->udma_four = 0;
	hwif->no_lba48 = 1;	/* FIXME ? Did some one test that ? */
	hwif->no_lba48_dma = 1;

	hwif->drives[0].unmask = 1;
	hwif->drives[0].autotune = 0; /* default */
	hwif->drives[0].autodma = hwif->autodma;
	hwif->drives[0].no_io_32bit = 1;	/* Anyone tried ? */

	hwif->drives[1].unmask = 1;
	hwif->drives[1].autotune = 0; /* default */
	hwif->drives[1].autodma = hwif->autodma;
	hwif->drives[1].no_io_32bit = 1;	/* Anyone tried ? */
 
	hwif->hwif_data = priv;
	dev_set_drvdata(dev, hwif);

	/* Lauch probe */
	probe_hwif_init(hwif);

	/* We're good ! */
	printk(KERN_INFO
	  "mpc52xx-ide: Setup successful for %s (mem=%08lx-%08lx irq=%d)\n",
	  hwif->name, res.start, res.end, ata_irq);

	return 0;

	/* Error path */
error:
	if (ata_regs)
		iounmap(ata_regs);
	release_mem_region(res.start, sizeof(struct mpc52xx_ata));
	if (priv) {
		kfree(priv);
	}
	return rv;
}

static int mpc52xx_ide_remove (struct device * dev) {
	ide_hwif_t *hwif = dev_get_drvdata(dev);
	struct mpc52xx_ide_priv *priv = hwif->hwif_data;

	ide_unregister(ide_hwifs - hwif);
	iounmap(priv->ata_regs);

	return 0;
}

#ifdef CONFIG_PM
static int
mpc52xx_ide_suspend(struct platform_device *dev, pm_message_t state)
{
	return 0;	/* FIXME : What to do here ? */
}

static int
mpc52xx_ide_resume(struct platform_device *dev)
{
	return 0;	/* FIXME : What to do here ? */
}
#endif

static struct device_driver mpc52xx_ide_driver = {
	.name	= mpc52xx_ide_string,
	.bus	= &platform_bus_type,
	.probe	= mpc52xx_ide_probe,
	.remove	= mpc52xx_ide_remove,
#ifdef CONFIG_PM
	.suspend	= mpc52xx_ide_suspend,
	.resume		= mpc52xx_ide_resume,
#endif
};

/* ======================================================================== */
/* Module                                                                   */
/* ======================================================================== */
static void mpc52xx_ide_platform_release(struct device *device)
{
	struct platform_device *pldev;

	/* free device */
	pldev = to_platform_device(device);
	kfree(pldev);
}

static int __devinit mpc52xx_ide_init_module(void)
{
	struct platform_device *pldev;
	int err;

	printk(KERN_INFO "FSL Open Firmware MPC52xx IDE driver\n");
	if (driver_register(&mpc52xx_ide_driver)) {
		printk(KERN_ERR "MPC52xx IDE Driver registration failed\n");
		err = -ENODEV;
		goto out;
	}

        if (!(pldev = kmalloc(sizeof (*pldev), GFP_KERNEL))) {
		err = -ENOMEM;
		goto out_unregister_driver;
	}

	memset (pldev, 0, sizeof (*pldev));
	pldev->name		= mpc52xx_ide_string;
	pldev->id		= 0;
	pldev->dev.release	= mpc52xx_ide_platform_release;

	if (platform_device_register(pldev)) {
		err = -ENODEV;
		goto out_free_pldev;
	}

        if (!pldev->dev.driver) {
		/*
		 * The driver was not bound to this device, there was
                 * no hardware at this address. Unregister it, as the
		 * release fuction will take care of freeing the
		 * allocated structure
		 */
		platform_device_unregister (pldev);
	}

	mpc52xx_ide_dev = pldev;
	return 0;

out_free_pldev:
	kfree(pldev);

out_unregister_driver:
	driver_unregister(&mpc52xx_ide_driver);
out:
	return err;
}

static void mpc52xx_ide_exit_module(void)
{
	driver_unregister(&mpc52xx_ide_driver);
	platform_device_unregister(mpc52xx_ide_dev);
}

module_init(mpc52xx_ide_init_module);
module_exit(mpc52xx_ide_exit_module);

MODULE_AUTHOR("Sylvain Munaut <tnt@246tNt.com>");
MODULE_DESCRIPTION("Freescale MPC52xx IDE/ATA driver");
MODULE_LICENSE("GPL");

