/*
 * drivers/misc/mv64460_timer.c
 *
 * marvell (mv64460) programmable timers API.
 *
 * Author:  Majid Khan M Z, (amm117@motorola.com).
 * Copyright 
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/types.h> 
#include <asm/uaccess.h>
#include <linux/interrupt.h>
#include <asm/mv64x60.h>
#include <asm/mv64x60_defs.h>
#include "mv64460_timer.h"

#define BIT0	0x00000001
#define BIT1	0x00000002
#define BIT2	0x00000004
#define BIT3	0x00000008

		/* Marvell timer control register for timer 0 */
#define MV64X60_TC_EN		BIT0
#define MV64X60_TC_MODE 	BIT1
#define MV64X60_TC_TRIG		BIT2
#define MV64X60_TC_WIDTH	BIT3
#define MV64X60_TC_RES		0xf0

#define DRV_NAME "mv64x60_timer"
#define DRV_VERSION "1.0"

static int mv64x60_timer_irq[]= {8, 9, 10, 11};

struct mv64x60_timer_t {
	mv64x60_timer_f func;
	void *func_data;
	int started;
	int periodic; /* 1 for periodic and 0 for one shot timer */
	unsigned long period;
};
static struct mv64x60_handle bh;
static struct mv64x60_timer_t mv64x60_timer[4];
static spinlock_t mv64x60_timer_lock = SPIN_LOCK_UNLOCKED;
static unsigned long mv64x60_bus_clk_tick = 133;	/*number of ticks per microseconds*/

int mv64x60_timer_stop(int timer_number)
{
	unsigned long flags;
	if(timer_number < 0 || timer_number > 3)
		return (-EINVAL);
	spin_lock_irqsave(&mv64x60_timer_lock, flags);
	mv64x60_clr_bits(&bh, MV64x60_TIMR_CNTR_0_3_CNTL, MV64X60_TC_EN << (timer_number * 8));
	mv64x60_timer[timer_number].func = NULL;
	mv64x60_timer[timer_number].func_data = NULL;
	mv64x60_timer[timer_number].started = 0;
	mv64x60_timer[timer_number].periodic = 0;
	mv64x60_timer[timer_number].period = 0;
	spin_unlock_irqrestore(&mv64x60_timer_lock, flags);
	return (0);
}
EXPORT_SYMBOL(mv64x60_timer_stop);

int mv64x60_timer_start(int timer_number, unsigned long microseconds, mv64x60_timer_f func, void *func_data, int periodic)
{
	unsigned long flags;
	
	printk(KERN_INFO "mv64x60_timer_start: number %d micoroseconds 0x%lx periodic %d\n", timer_number, microseconds, periodic);
	
	if(timer_number < 0 || timer_number > 3)
		return (-EINVAL);
	if(func == NULL)
		return (-EINVAL);
	if(mv64x60_timer[timer_number].started == 1)
		return (-EBUSY);
	
	if(microseconds > (0xffffffff/mv64x60_bus_clk_tick) || microseconds <= 0)
		return (-EINVAL);
	spin_lock_irqsave(&mv64x60_timer_lock, flags);

	mv64x60_timer[timer_number].started = 1;
       	mv64x60_timer[timer_number].func = func;
	mv64x60_timer[timer_number].func_data = func_data;
       	mv64x60_timer[timer_number].periodic = periodic;
       	mv64x60_timer[timer_number].period = microseconds;

	mv64x60_write(&bh, MV64x60_TIMR_CNTR_0 + (timer_number * 4), (microseconds * mv64x60_bus_clk_tick));
	mv64x60_set_bits(&bh, MV64x60_TIMR_CNTR_0_3_CNTL, MV64X60_TC_EN << (timer_number * 8)); /* enable the timer */

	spin_unlock_irqrestore(&mv64x60_timer_lock, flags);
	return (0);
	
}
EXPORT_SYMBOL(mv64x60_timer_start);

static irqreturn_t mv64x60_timer_int_handler (int irq, void *data, struct pt_regs *regs)
{
	struct mv64x60_timer_t *info = data;
	mv64x60_timer_f func;
	void *func_data;
	int timer_number;
	unsigned int caused_val;

	caused_val = mv64x60_read(&bh, MV64x60_TIMR_CNTR_0_3_INTR_CAUSE) & 0xf;
	switch(caused_val) {
		case BIT0:
			timer_number = 0;
			info = &mv64x60_timer[timer_number];
			func_data = info->func_data;
			func = info->func;
			mv64x60_clr_bits(&bh, MV64x60_TIMR_CNTR_0_3_INTR_CAUSE, BIT0);
			if(info->func)
				func(func_data);
			if(!info->periodic) 
				mv64x60_timer_stop(timer_number);
			break;
		case BIT1:
			timer_number = 1;
			info = &mv64x60_timer[timer_number];
			func_data = info->func_data;
			func = info->func;
			mv64x60_clr_bits(&bh, MV64x60_TIMR_CNTR_0_3_INTR_CAUSE, BIT1);
			if(info->func)
				func(func_data);
			if(!info->periodic) 
				mv64x60_timer_stop(timer_number);
			break;
		case BIT2:
			timer_number = 2;
			info = &mv64x60_timer[timer_number];
			func_data = info->func_data;
			func = info->func;
			mv64x60_clr_bits(&bh, MV64x60_TIMR_CNTR_0_3_INTR_CAUSE, BIT2);
			if(info->func)
				func(func_data);
			if(!info->periodic)
				mv64x60_timer_stop(timer_number);
			break;
		case BIT3:
			timer_number = 3;
			info = &mv64x60_timer[timer_number];
			func_data = info->func_data;
			func = info->func;
			mv64x60_clr_bits(&bh, MV64x60_TIMR_CNTR_0_3_INTR_CAUSE, BIT3);
			if(info->func)
				func(func_data);
			if(!info->periodic)
				mv64x60_timer_stop(timer_number);
			break;
		default:
			return (IRQ_NONE);
	}
	return IRQ_HANDLED;
}

#ifdef CONFIG_PROC_FS
static int mv64x60_timer_proc_read(char *buf, char **start, off_t off, int len, int *eof, void *data)
{
	int i;
	len = 0;
	len += sprintf(buf+len, "\nTimer	Active	periodic	period(microseconds)\n");
	len += sprintf(buf+len,   "_____	______	________	_____________________");
	for(i = 0; i <= 3; i++) {
		len += sprintf(len+buf, "\n#%d	%c	%c	%ld\n", i+1, (mv64x60_timer[i].started) ? 'y' : 'n', 
				(mv64x60_timer[i].periodic) ? 'y' : 'n', (mv64x60_timer[i].periodic == 1) ? mv64x60_timer[i].period : 0);
	}
	return (len);
}
#endif

static __init int mv64x60_timer_init(void)
{
	int i;
	printk(KERN_INFO DRV_NAME " version " DRV_VERSION "\n");

	bh.v_base = mv64x60_get_bridge_vbase();

	mv64x60_write(&bh, MV64x60_TIMR_CNTR_0_3_CNTL, 0); /* set to default */
	for(i = 0; i <= 3; i++) {
		mv64x60_timer[i].func = NULL;
		mv64x60_timer[i].func_data = NULL;
		mv64x60_timer[i].started = 0;
		mv64x60_timer[i].periodic = 0;
		mv64x60_timer[i].period = 0;
		mv64x60_set_bits(&bh, MV64x60_TIMR_CNTR_0_3_CNTL, MV64X60_TC_MODE << (i * 8)); /* set to timer mode */
		mv64x60_set_bits(&bh, MV64x60_TIMR_CNTR_0_3_INTR_MASK, 1 << i); /* enable the timer interrupt */
	}
	
	for (i = 0; i <= 3; i++)
		if(request_irq(mv64x60_timer_irq[i], mv64x60_timer_int_handler, SA_INTERRUPT | SA_SHIRQ, "MV64x60_Timer", &mv64x60_timer[i])) {
        	        printk("can't register irq:\n");
                	return (-EIO);
        	}
#ifdef CONFIG_PROC_FS
	create_proc_read_entry("mv64x60_timers", 0, NULL, mv64x60_timer_proc_read, NULL);
#endif
	return 0;
}
static __exit void mv64x60_timer_exit(void)
{
	int i;
	for(i = 0; i <= 3; i++) {
		mv64x60_timer_stop(i);
		free_irq(mv64x60_timer_irq[i], &mv64x60_timer[i]);
	}
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Majid Khan M Z <amm117@motorola.com>");
MODULE_DESCRIPTION("Marvell(MV64x60) timer driver");

module_init(mv64x60_timer_init);
module_exit(mv64x60_timer_exit);

