/*
 * FILE NAME : 280fbs.c
 *
 * Mapping for Intel Strataflash on the Motorola Force PrPMC280
 *
 * Joseph Barnett <jbarnett@motorola.com>
 *
 * Copyright 2004 Motorola Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/concat.h>
#include <linux/mtd/partitions.h>
#include <platforms/ppmc280.h>
#include <platforms/ppmc280_spd.h>

extern struct ppmc280_board_def ppmc280_definition;

#define PPMC280_CS0_NAME    "PPMC280 User Flash Bank 0"
#define PPMC280_CS1_NAME    "PPMC280 User Flash Bank 1"

#define BANK_CONCATINATION_SIZE (PPMC280_EXT_FLASH_SIZE_ACTUAL * 2)

#define FW_IMG_SIZE              0x00100000
#define FW_IMG_CFG_DATA_SIZE     0x00040000

#define FW_IMG_BANK_A_OFFSET     0x00000000
#define FW_IMG_BANK_B_OFFSET    (BANK_CONCATINATION_SIZE - FW_IMG_SIZE)
#define FW_IMG_CFG_DATA_OFFSET   FW_IMG_SIZE

#define OS_KERN_IMG_OFFSET      (FW_IMG_CFG_DATA_OFFSET + FW_IMG_CFG_DATA_SIZE)
#define OS_KERN_IMG_SIZE         0x00400000

#define OS_FILESYS_OFFSET       (OS_KERN_IMG_OFFSET + OS_KERN_IMG_SIZE)
#define OS_FILESYS_SIZE         (FW_IMG_BANK_B_OFFSET - OS_FILESYS_OFFSET)

/* #define FBS280_DEBUG 1 */
#define PFX "280FBS: "

static struct map_info ppmc280_r2_map[] = { 
     { 
         .name = PPMC280_CS0_NAME, 
         .size = PPMC280_EXT_FLASH_SIZE_ACTUAL_R2, 
         .bankwidth = 4, 
         .phys = PPMC280_EXT_FLASH_BASE_1 
     },
}; 

static struct map_info ppmc280_map[] = {
     {
         .name = PPMC280_CS0_NAME,
         .size = PPMC280_EXT_FLASH_SIZE_ACTUAL,
         .bankwidth = 4,
         .phys = PPMC280_EXT_FLASH_BASE_1
     },
     {
         .name = PPMC280_CS1_NAME,
         .size = PPMC280_EXT_FLASH_SIZE_ACTUAL,
         .bankwidth = 4,
         .phys = PPMC280_EXT_FLASH_BASE_2
     },
};

#define NUM_FLASH_BANKS	(sizeof(ppmc280_map)/sizeof(struct map_info))
#define NUM_FLASH_BANKS_R2 (sizeof(ppmc280_r2_map)/sizeof(struct map_info)) 
#define MAX_NUM_FLASH_BANKS (MAX(NUM_FLASH_BANKS, NUM_FLASH_BANKS_R2))

static struct mtd_partition ppmc280_part[] = {
    {
        .name       = "Firmware Image A",
        .offset     = FW_IMG_BANK_A_OFFSET,
        .size       = FW_IMG_SIZE,
    },
    {
        .name       = "Firmware Configuration Data",
        .offset     = FW_IMG_CFG_DATA_OFFSET,
        .size       = FW_IMG_CFG_DATA_SIZE,
#ifdef CONFIG_MTD_PPMC280_FORCE_READONLY_FIRMWARE
        .mask_flags = MTD_WRITEABLE,
#endif
    },
    {
        .name       = "Embedded OS Kernel Image",
        .offset     = OS_KERN_IMG_OFFSET,
        .size       = OS_KERN_IMG_SIZE,
    },
    {
        .name       = "Embedded OS File System",
        .offset     = OS_FILESYS_OFFSET,
        .size       = OS_FILESYS_SIZE,
    },
    {
        .name       = "Firmware Image B",
        .offset     = FW_IMG_BANK_B_OFFSET,
        .size       = FW_IMG_SIZE,
    }
};

static struct mtd_info *flash[MAX_NUM_FLASH_BANKS];
static struct mtd_info *chunk;
static struct map_info *map;

static int __init init_ppmc280(void)
{
    int                dev_cnt = 0,
                       num_flash_banks = 0,
                       i       = 0;

	if (ppmc280_definition.pcb_rev == PPMC280_REV_NONROHS) {
		num_flash_banks = NUM_FLASH_BANKS;
		map = ppmc280_map;
#ifdef FBS280_DEBUG
		printk(KERN_NOTICE PFX "Board rev %x is Non-ROHS\n",
				ppmc280_definition.pcb_rev); 
#endif /* FBS280_DEBUG */
	} else {
		num_flash_banks = NUM_FLASH_BANKS_R2;
		map = ppmc280_r2_map;
#ifdef FBS280_DEBUG
		printk(KERN_NOTICE PFX "Board rev %x is ROHS\n",
				ppmc280_definition.pcb_rev); 
#endif /* FBS280_DEBUG */
	}
#ifdef FBS280_DEBUG
	printk(KERN_NOTICE PFX "# of Flash banks: %x\n", num_flash_banks);
#endif /* FBS280_DEBUG */
  
    for (i = 0; i < num_flash_banks; i++) { 
        printk(KERN_NOTICE "PPMC280 flash device: 0x%lx at 0x%llx\n", map[i].size, (unsigned long long) map[i].phys); 
        map[i].virt = ioremap_nocache(map[i].phys, map[i].size); 

        if (!map[i].virt) { 
            printk("Failed to ioremap_nocache\n");
            return -EIO;
        }

        simple_map_init(&map[i]); 

        flash[i]     = do_map_probe("cfi_probe",   &map[i]); 
        if(!flash[i])
            flash[i] = do_map_probe("jedec_probe", &map[i]); 
        if(!flash[i])
            flash[i] = do_map_probe("map_rom",     &map[i]); 

        if (flash[i]) {
            flash[i]->owner = THIS_MODULE;
            ++dev_cnt;
        } else {
            iounmap((void *)map[i].virt); 
        }
    }

    chunk=0;
    if(dev_cnt >= 2) {
        /* Combine the two flash banks into a single MTD device & register it: */
        if(NULL != (chunk = mtd_concat_create(flash, 2, "PPMC280 Flash Banks 0 and 1"))) {
            add_mtd_device(chunk);
            add_mtd_partitions(chunk, ppmc280_part, ARRAY_SIZE(ppmc280_part));
#ifdef CONFIG_PPMC280_MAP_USERFLASH_READONLY
            printk("PPMC280 user flash was mapped read-only\n");
#endif
#ifdef CONFIG_MTD_PPMC280_FORCE_READONLY_FIRMWARE
            printk("PPMC280 firmware configuration data partition was set read-only\n");
#endif
        }
    } else {   
	if(dev_cnt > 0) {
            add_mtd_partitions(flash[0], ppmc280_part, ARRAY_SIZE(ppmc280_part)); 
        } else {
	  return(-ENXIO); 
        } 
    } 
    return(0);
}

static void __exit cleanup_ppmc280(void)
{
    int i;
    int num_flash_banks=0;
    struct map_info *map;

    if (ppmc280_definition.pcb_rev == PPMC280_REV_NONROHS) {
        map = ppmc280_map;
        num_flash_banks = NUM_FLASH_BANKS;
    } else {
        map = ppmc280_r2_map;
        num_flash_banks = NUM_FLASH_BANKS_R2;
    }
    if (chunk) {
        del_mtd_partitions(chunk);
        del_mtd_device(chunk);
        mtd_concat_destroy(chunk);
    } else {
        del_mtd_partitions(flash[0]);
    }

    for (i = 0; i < num_flash_banks; i++) {  /* NUM_FLASH_BANKS is the Maximum */
        if (flash[i])
            map_destroy(flash[i]);

        if (map[i].virt) { 
            iounmap((void *)map[i].virt); 
            map[i].virt = 0; 
        }
    }
}

module_init(init_ppmc280);
module_exit(cleanup_ppmc280);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Joseph Barnett <jbarnett@motorola.com>");
MODULE_DESCRIPTION("Flash Support for the Motorola ATCA PPMC280");
