/*
 * Flash mapping for BCM91125 SDK boards
 *
 * Copyright (C) 2004 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#define WINDOW_ADDR 0x1fc00000
#define WINDOW_SIZE 0x1000000
#define BUSWIDTH 1

static struct mtd_info *bcm91125_mtd;

struct map_info bcm91125_map = {
	.name = "Physically mapped flash",
	.size = WINDOW_SIZE,
	.bankwidth = BUSWIDTH,
	.phys = WINDOW_ADDR,
};

#ifdef CONFIG_MTD_PARTITIONS

static struct mtd_partition bcm91125_mtd_parts[] __initdata = {
	{
		.name = "boot",
		.size = 0x0e0000,
		.offset = 0x000000,
		.mask_flags = MTD_WRITEABLE
	}, {
		.name = "nvram",
		.size = 0x020000,
		.offset = 0x0e0000
	}, {
		.name = "os",
		.size = 0x700000,
		.offset = 0x100000
	}, {
		.name = "user",
		.size = 0x800000,
		.offset = 0x700000
	}
};

#define NUM_PARTS (sizeof(bcm91125_mtd_parts)/sizeof(struct mtd_partition))

#define PTABLE_MAGIC            0x5054424C /* 'PTBL' (big-endian) */
#define PTABLE_MAGIC_LE         0x4C425450 /* 'PTBL' (little-endian) */
#define PTABLE_PNAME_MAX        16
#define PTABLE_MAX_PARTITIONS   8

typedef struct partition_s {
    char        name[PTABLE_PNAME_MAX];
    uint32_t    offset;
    uint32_t    size;
    uint32_t    type;
    uint32_t    flags;
} partition_t;

typedef struct ptable_s {
    uint32_t    magic;
    uint32_t    version;
    uint32_t    chksum;
    uint32_t    reserved;
    partition_t part[PTABLE_MAX_PARTITIONS];
} ptable_t;

static struct mtd_partition bcm91125_mtd_user_parts[PTABLE_MAX_PARTITIONS];
static ptable_t ptable;
static int num_parts;

static int bcm91125_check_ptable(ptable_t *ptbl,
				 struct mtd_partition *part, 
				 ssize_t max_parts, ssize_t *num_parts)
{
	uint32_t chksum, *p32;
	int i, swapped = 0;

	if (ptbl->magic == PTABLE_MAGIC_LE) {
		swapped = 1;
	}
	else if (ptbl->magic != PTABLE_MAGIC) {
		return -1;
	}
	chksum = 0;
	p32 = (uint32_t*)ptbl;
	for (i = 0; i < sizeof(ptable_t)/4; i++) {
		chksum ^= p32[i];
		if (swapped) {
			p32[i] = swab32(p32[i]);
		}
	}
	if (chksum != 0) {
		return -1;
	}
	for (i = 0; i < max_parts && ptbl->part[i].size; i++) {
		part[i].name = ptbl->part[i].name;
		part[i].size = ptbl->part[i].size;
		part[i].offset = ptbl->part[i].offset;
	}
	*num_parts = i;
	return 0;
}

static int bcm91125_find_partitions(struct mtd_info *mi, 
				     struct mtd_partition *part, 
				     ssize_t max_parts, ssize_t *num_parts)
{
	int retlen, offset;

	if (mi == NULL || mi->read == NULL) {
		return -1;
	}
	for (offset = 0; offset < mi->size; offset += 0x8000) {
		if (mi->read(mi, offset, sizeof(ptable), &retlen, 
			     (u_char *)&ptable) < 0) {
			return -1;
		}
		if (retlen != sizeof(ptable)) {
			return -1;
		}
		if (bcm91125_check_ptable(&ptable, part, max_parts, 
					  num_parts) == 0) {
			if (*num_parts > 0) {
				printk("Found flash partition table "
				       "at offset 0x%08x\n", offset);
				return 0;
			}
		}
	}
	printk("No valid flash partition table found - using default mapping\n");
	return -1;
}

#endif

int __init init_bcm91125_map(void)
{
	int ret = 0;

	bcm91125_map.virt = ioremap(WINDOW_ADDR, WINDOW_SIZE);
	if (!bcm91125_map.virt) {
		printk(KERN_ERR "pflash: ioremap failed\n");
		ret = -EIO;
		goto fail;
	}
	simple_map_init(&bcm91125_map);

	if (!(bcm91125_mtd = do_map_probe("cfi_probe", &bcm91125_map))) {
		printk(KERN_ERR "pflash: cfi_probe failed\n");
		ret = -ENXIO;
		goto fail;
	}

	bcm91125_mtd->owner = THIS_MODULE;

	printk(KERN_NOTICE "Flash device: 0x%x at 0x%x\n", bcm91125_mtd->size, WINDOW_ADDR);

#ifdef CONFIG_MTD_PARTITIONS
	if (bcm91125_find_partitions(bcm91125_mtd, bcm91125_mtd_user_parts, 
				     PTABLE_MAX_PARTITIONS, &num_parts) == 0) {
		ret = add_mtd_partitions(bcm91125_mtd, bcm91125_mtd_user_parts, num_parts);
	}
	else {
		ret = add_mtd_partitions(bcm91125_mtd, bcm91125_mtd_parts, NUM_PARTS);
	}
	if (ret) {
		printk(KERN_ERR "pflash: add_mtd_partitions failed\n");
		goto fail;
	}
#endif

	return 0;

 fail:
	if (bcm91125_mtd)
		map_destroy(bcm91125_mtd);
	if (bcm91125_map.map_priv_1)
		iounmap((void *) bcm91125_map.map_priv_1);
	bcm91125_map.map_priv_1 = 0;
	return ret;
}

static void __exit cleanup_bcm91125_map(void)
{
#ifdef CONFIG_MTD_PARTITIONS
	del_mtd_partitions(bcm91125_mtd);
#endif
	map_destroy(bcm91125_mtd);
	iounmap((void *) bcm91125_map.map_priv_1);
	bcm91125_map.map_priv_1 = 0;
}

module_init(init_bcm91125_map);
module_exit(cleanup_bcm91125_map);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Flash mapping for BCM91125 SDK boards");
