/*
 * pramc7201_bflash.c
 *
 * The map file for Motorola PrAMC-7201 Boot Flash.
 *
 * Author: Motorola
 *         Harish K <harish@motorola.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 *  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>


#define FLASH_PHYS_ADDR 0xFFE00000	/* Boot Flash Area */

#define FLASH_SIZE 0x200000  

/*
 * Two partitions have been defined so far:
 *
 * 1. Partition 1 (0xFFE00000 - 0xFFEFFFFF)  -- 1.00MB
 * 2. Partition 2 (0xFFF00000 - 0xFFFFFFFF)  -- 1.00MB
 */

map_word pramc7201_bflash_read8(struct map_info *map, unsigned long ofs)
{
	map_word val;
    val.x[0] = (__u8) __raw_readb((void __iomem *) (map->map_priv_1 + ofs));
	return val;
}

map_word pramc7201_bflash_read16(struct map_info *map, unsigned long ofs)
{
	map_word val;
    val.x[0] = (__u16) __raw_readw((void __iomem *) (map->map_priv_1 + ofs));
    return val;
}

map_word pramc7201_bflash_read32(struct map_info *map, unsigned long ofs)
{
	map_word val;
    val.x[0] = (__u32) __raw_readl((void __iomem *) (map->map_priv_1 + ofs));
    return val;
}

void pramc7201_bflash_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void __iomem *) (map->map_priv_1 + from), len);
}

void pramc7201_bflash_write8(struct map_info *map, map_word val, unsigned long adr)
{
	__raw_writeb((__u8) val.x[0], (void __iomem *) (map->map_priv_1 + adr));
	mb();
}

void pramc7201_bflash_write16(struct map_info *map, map_word val, unsigned long adr)
{
	__raw_writew((__u16) val.x[0], (void __iomem *) (map->map_priv_1 + adr));
	mb();
}

void pramc7201_bflash_write32(struct map_info *map, map_word val, unsigned long adr)
{
	__raw_writel((__u32) val.x[0], (void __iomem *) (map->map_priv_1 + adr));
	mb();
}

void pramc7201_bflash_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void __iomem *) (map->map_priv_1 + to), from, len);
}

struct map_info pramc7201_bflash_map = {
	.name = "Boot flash device",
	.phys = FLASH_PHYS_ADDR,
	.size = FLASH_SIZE,
	.bankwidth = 1,
	.read = pramc7201_bflash_read8,
	.copy_from = pramc7201_bflash_copy_from,
	.write = pramc7201_bflash_write8,
	.copy_to = pramc7201_bflash_copy_to
};

struct mtd_partition pramc7201_bflash_parts[] = {
        {
                name    :               "Bank 1",
                offset  :               0x0,
                size    :               0x100000,
        },{
                name    :               "Bank 2",
                offset  :               0x100000,
                size    :               0x100000
        }
};


#define PARTITION_COUNT (sizeof(pramc7201_bflash_parts)/sizeof(struct mtd_partition))

static struct mtd_info *mymtd;

#define	T_BLOCK_MINUS15		0xFFA00002

void fwh_init (void) 
{
	unsigned char *fwh_tblocks;
	unsigned int count=0;

	/*
	** set sectors to be in full access lock down.  This means that the
	** value of the lock bit cannot be changed, i.e. the chip is always
	** configured to allow read and write access.
	*/

	for (count=0; count<=0x1F0000; count+=0x10000) {
		fwh_tblocks = (unsigned char*)ioremap_nocache ((T_BLOCK_MINUS15+count),4);
		*fwh_tblocks = 0x02;
		iounmap ((unsigned char*)fwh_tblocks);
	}
}

static struct pci_device_id pci_tbl[] =
{
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ESB_1,
	  PCI_ANY_ID, PCI_ANY_ID, },
	{ 0, },
};

static struct pci_dev		*lpc_pdev = NULL;

#define BIOS_CNTL	0x4e

int __init init_pramc7201_bflash (void)
{	
	struct pci_device_id	*id;
	u16			word;

	lpc_pdev = NULL;
	for (id = pci_tbl; id->vendor; id++)
	{
		lpc_pdev = pci_find_device(id->vendor, id->device, NULL);
		if (lpc_pdev)
		{
			break;
		}
	}
	if (!lpc_pdev)
	{
		printk(KERN_ERR "Unable to locate LPC interface controller\n");
		return(-ENODEV);
	}

	/* Enable writes through the rom window */
	pci_read_config_word(lpc_pdev, BIOS_CNTL, &word);
	if (!(word & 1)  && (word & (1<<1))) {
		/* The BIOS will generate an error if I enable
		 * this device, so don't even try.
		 */
		printk(KERN_ERR "Unable to enable firmware writes\n");
		return(-ENODEV);
	}
	pci_write_config_word(lpc_pdev, BIOS_CNTL, word | 1);

	/*unlock blocks*/
	fwh_init ();

	printk(KERN_NOTICE "Boot flash device: %x at %x\n",
			FLASH_SIZE, FLASH_PHYS_ADDR);
	
	pramc7201_bflash_map.map_priv_1 = (unsigned long)ioremap( FLASH_PHYS_ADDR, FLASH_SIZE ) ;

	if (!pramc7201_bflash_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	mymtd = do_map_probe("jedec_probe", &pramc7201_bflash_map);
	if (mymtd) {
		mymtd->owner = THIS_MODULE;

		/*
		** since the chip cannot be locked make sure the functions are
		** not configured so that -EOPNOTSUPP is returned when user
		** attempts to lock an area
		*/
		mymtd->lock = NULL;
		mymtd->unlock = NULL;
		add_mtd_partitions(mymtd, pramc7201_bflash_parts, PARTITION_COUNT);
		printk(KERN_NOTICE "Boot flash device initialized\n");
		return 0;
	}

	iounmap((void *)pramc7201_bflash_map.map_priv_1);

	return -ENXIO;
}

static void __exit cleanup_pramc7201_bflash (void)
{
	u16			word;

	if (mymtd) {
		del_mtd_partitions(mymtd);
		map_destroy(mymtd);
	}
	if (pramc7201_bflash_map.map_priv_1) {
		iounmap((void *)pramc7201_bflash_map.map_priv_1);
		pramc7201_bflash_map.map_priv_1 = 0;
	}
	if (lpc_pdev)
	{
		/* Disable writes through the rom window */
		pci_read_config_word(lpc_pdev, BIOS_CNTL, &word);
		pci_write_config_word(lpc_pdev, BIOS_CNTL, word & ~1);
		lpc_pdev = NULL;
	}
}

module_init(init_pramc7201_bflash);
module_exit(cleanup_pramc7201_bflash);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Harish K <harish@motorola.com>");
MODULE_DESCRIPTION("MTD map driver for PrAMC-7201 Boot  Flash chip");
