/*
 * Driver for the MPC5200 Fast Ethernet Controller - PHY/MII part
 *
 * Copyright (C) 2007  Domen Puncer, Telargo, Inc.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2. This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/phy.h>
#include <asm/io.h>
#include <asm/mpc52xx.h>
#include <asm/of_platform.h>
#include "fec.h"

struct fec_mdio_priv {
	int completed;
	wait_queue_head_t wq;
	struct mpc52xx_fec __iomem *regs;
	int irq;
};

static int fec_mdio_read(struct mii_bus *bus, int phy_id, int reg)
{
	struct fec_mdio_priv *priv = bus->priv;
	int tries = 100;

	u32 request = FEC_MII_READ_FRAME;
	request |= (phy_id << FEC_MII_DATA_PA_SHIFT) & FEC_MII_DATA_PA_MSK;
	request |= (reg << FEC_MII_DATA_RA_SHIFT) & FEC_MII_DATA_RA_MSK;

	out_be32(&priv->regs->mii_data, request);

	/* wait for it to finish, this takes about 23 us on lite5200b */
	while (priv->completed == 0 && tries--)
		udelay(5);

	priv->completed = 0;

	if (tries == 0)
		return -ETIMEDOUT;

	return in_be32(&priv->regs->mii_data) & FEC_MII_DATA_DATAMSK;
}

static int fec_mdio_write(struct mii_bus *bus, int phy_id, int reg, u16 data)
{
	struct fec_mdio_priv *priv = bus->priv;
	u32 value = data;
	int tries = 100;

	value |= FEC_MII_WRITE_FRAME;
	value |= (phy_id << FEC_MII_DATA_PA_SHIFT) & FEC_MII_DATA_PA_MSK;
	value |= (reg << FEC_MII_DATA_RA_SHIFT) & FEC_MII_DATA_RA_MSK;

	out_be32(&priv->regs->mii_data, value);

	/* wait for request to finish */
	while (priv->completed == 0 && tries--)
		udelay(5);

	priv->completed = 0;

	if (tries == 0)
		return -ETIMEDOUT;

	return 0;
}

static irqreturn_t fec_mdio_interrupt(int irq, void *dev_id)
{
	struct fec_mdio_priv *priv = dev_id;
	struct mpc52xx_fec __iomem *fec;
	int ievent;

	fec = priv->regs;
	ievent = in_be32(&fec->ievent);

	ievent &= FEC_IEVENT_MII;
	if (!ievent)
		return IRQ_NONE;

	out_be32(&fec->ievent, ievent);

	priv->completed = 1;
	wake_up(&priv->wq);

	return IRQ_HANDLED;
}

static int fec_mdio_probe(struct of_device *of, const struct of_device_id *match)
{
	struct device *dev = &of->dev;
	struct device_node *np = of->node;
	struct device_node *child = NULL;
	struct mii_bus *bus;
	struct fec_mdio_priv *priv;
	struct resource res = {};
	int err;
	int i;

	bus = kzalloc(sizeof(*bus), GFP_KERNEL);
	if (bus == NULL)
		return -ENOMEM;
	priv = kzalloc(sizeof(*priv), GFP_KERNEL);
	if (priv == NULL) {
		err = -ENOMEM;
		goto out_free;
	}

	bus->name = "mpc52xx MII bus";
	bus->read = fec_mdio_read;
	bus->write = fec_mdio_write;

	/* setup irqs */
	bus->irq = kmalloc(sizeof(bus->irq[0]) * PHY_MAX_ADDR, GFP_KERNEL);
	if (bus->irq == NULL) {
		err = -ENOMEM;
		goto out_free;
	}
	for (i=0; i<PHY_MAX_ADDR; i++)
		bus->irq[i] = PHY_POLL;

	while ((child = of_get_next_child(np, child)) != NULL) {
		int irq = irq_of_parse_and_map(child, 0);
		if (irq != NO_IRQ) {
			const u32 *id = get_property(child, "reg", NULL);
			bus->irq[*id] = irq;
		}
	}

	/* setup registers */
	err = of_address_to_resource(np, 0, &res);
	if (err)
		goto out_free;
	priv->regs = ioremap(res.start, res.end - res.start + 1);
	if (priv->regs == NULL) {
		err = -ENOMEM;
		goto out_free;
	}

	priv->irq = irq_of_parse_and_map(np, 0);
	err = request_irq(priv->irq, &fec_mdio_interrupt, IRQF_DISABLED | IRQF_SHARED,
	                "fec_mdio", priv);
	if (err) {
		printk(KERN_ERR "%s: interrupt request failed with %i\n", __func__, err);
		goto out_unmap;
	}

	bus->id = res.start;
	bus->priv = priv;

	bus->dev = dev;
	dev_set_drvdata(dev, bus);

	init_waitqueue_head(&priv->wq);

	/* set MII speed */
	out_be32(&priv->regs->mii_speed, ((mpc52xx_find_ipb_freq(of->node) >> 20) / 5) << 1);

	/* enable MII interrupt */
	out_be32(&priv->regs->imask, in_be32(&priv->regs->imask) | FEC_IMASK_MII);

	err = mdiobus_register(bus);
	if (err)
		goto out_free_irq;

	return 0;

 out_free_irq:
	free_irq(priv->irq, dev);
	irq_dispose_mapping(priv->irq);
 out_unmap:
	iounmap(priv->regs);
 out_free:
	for (i=0; i<PHY_MAX_ADDR; i++)
		if (bus->irq[i] != PHY_POLL)
			irq_dispose_mapping(bus->irq[i]);
	kfree(bus->irq);
	kfree(priv);
	kfree(bus);

	return err;
}

static int fec_mdio_remove(struct of_device *of)
{
	struct device *dev = &of->dev;
	struct mii_bus *bus = dev_get_drvdata(dev);
	struct fec_mdio_priv *priv = bus->priv;
	int i;

	mdiobus_unregister(bus);
	dev_set_drvdata(dev, NULL);

	free_irq(priv->irq, dev);
	irq_dispose_mapping(priv->irq);
	iounmap(priv->regs);
	for (i=0; i<PHY_MAX_ADDR; i++)
		if (bus->irq[i])
			irq_dispose_mapping(bus->irq[i]);
	kfree(priv);
	kfree(bus->irq);
	kfree(bus);

	return 0;
}


static struct of_device_id fec_mdio_match[] = {
	{
		.type = "mdio",
		.compatible = "mpc5200b-fec-phy",
	},
	{},
};

static struct of_platform_driver fec_mdio_driver = {
	.name = "mpc5200b-fec-phy",
	.probe = fec_mdio_probe,
	.remove = fec_mdio_remove,
	.match_table = fec_mdio_match,
};


int __init fec_mdio_init(void)
{
	return of_register_platform_driver(&fec_mdio_driver);
}

void __exit fec_mdio_exit(void)
{
	of_unregister_platform_driver(&fec_mdio_driver);
}
